import asyncio
import typing

from .interfaces.controller import Controller
from .interfaces.actor import Actor


def schedule_pipelines(event_loop: asyncio.AbstractEventLoop, *leaves: Controller):
    seen = {None}  # type: typing.Set[typing.Optional[Controller]]
    for leave in leaves:
        this = leave
        while this not in seen:
            seen.add(this)
            if isinstance(this, Actor):
                this.mount(event_loop)
            this = getattr(this, 'target', None)  # type: typing.Optional[Controller]



import asyncio

from cobald.interfaces.pool import Pool
from cobald.interfaces.proxy import ProxyPool
from cobald.interfaces.actor import Actor


class Buffer(ProxyPool, Actor):
    """
    A timed buffer for changes to a pool

    :param target: the pool to which changes are applied
    :param window: interval after which changes are applied

    Any changes made to :py:attr:`demand` are stored internally.
    Every ``window`` seconds, the final demand is applied to ``target``.
    """
    demand = 0.0

    def __init__(self, target: Pool, window: float = 10.):
        super().__init__(target=target)
        self.window = window
        self.demand = target.demand

    @asyncio.coroutine
    def run(self):
        while True:
            self.target.demand = self.demand
            self.demand = self.target.demand
            yield from asyncio.sleep(self.window)



import time
import logging
import subprocess


def query_limits(query_command, key_transform):
    resource_limits = {}
    for item in subprocess.check_output(query_command, universal_newlines=True).splitlines():
        if '=' not in item:
            continue
        key, _, value = (value.strip() for value in item.partition('='))
        try:
            resource = key_transform(key)
        except ValueError:
            continue
        else:
            try:
                resource_limits[resource] = float(value)
            except ValueError:
                pass
    return resource_limits


class ConcurrencyConstraintView(object):
    def __init__(self, pool: str = None, max_age: float = 30.):
        self._logger = logging.getLogger('condor_limits.constraints')
        self.pool = pool
        self.max_age = max_age
        self._valid_date = 0
        self._constraints = {}

    def __getitem__(self, resource: str) -> float:
        if self._valid_date < time.time():
            self._query_constraints()
        try:
            return self._constraints[resource]
        except KeyError:
            if '.' in resource:
                return self._constraints[resource.split('.')[0]]  # check parent group of resource
            raise

    def __setitem__(self, key: str, value: float):
        self._set_constraint(key, value)

    @staticmethod
    def _key_to_resource(key: str) -> str:
        if key.lower()[-6:] == '_limit':
            return key[:-6]
        raise ValueError

    def _query_constraints(self):
        query_command = ['condor_config_val', '-negotiator', '-dump', 'LIMIT']
        if self.pool:
            query_command.extend(('-pool', str(self.pool)))
        resource_limits = query_limits(query_command, key_transform=self._key_to_resource)
        self._valid_date = self.max_age + time.time()
        self._constraints = resource_limits

    def _set_constraint(self, resource: str, constraint: float):
        reconfig_command = ['condor_config_val', '-negotiator']
        if self.pool:
            reconfig_command.extend(('-pool', str(self.pool)))
        reconfig_command.extend(('-rset', '%s_LIMIT = %s' % (resource, int(constraint))))
        try:
            subprocess.check_call(reconfig_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=10)
        except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as err:
            self._logger.error('failed to constraint %r to %r', resource, constraint, exc_info=err)
        else:
            self._constraints[resource] = constraint


class ConcurrencyUsageView(object):
    def __init__(self, pool: str = None, max_age: float = 30.):
        self.pool = pool
        self.max_age = max_age
        self._valid_date = 0
        self._usage = {}

    def __getitem__(self, resource: str) -> float:
        if self._valid_date < time.time():
            self._query_usage()
        try:
            return self._usage[resource.replace('.', '_')]
        except KeyError:
            if '.' in resource:
                return self._usage[resource.split('.')[0]]  # check parent group of resource
            raise

    @staticmethod
    def _key_to_resource(key: str) -> str:
        if key.startswith('ConcurrencyLimit_'):
            return key[17:]
        raise ValueError

    def _query_usage(self):
        query_command = ['condor_userprio', '-negotiator', '-long']
        if self.pool:
            query_command.extend(('-pool', str(self.pool)))
        resource_usage = query_limits(query_command, key_transform=self._key_to_resource)
        self._valid_date = self.max_age + time.time()
        self._usage = resource_usage

    def __str__(self):
        if self._valid_date < time.time():
            self._query_usage()
        return str(self._usage)

    def __repr__(self):
        return '%s(pool=%s, max_age=%s)' % (self.__class__.__name__, self.pool, self.max_age)


class PoolResources(object):
    def __init__(self, pool: str = None, max_age: float = 30.):
        self.pool = pool
        self.max_age = max_age
        self._valid_date = 0
        self._data = {}

    def __getitem__(self, resource: str) -> float:
        if self._valid_date < time.time():
            self._query_data()
        return self._data[resource]

    def _query_data(self):
        query_command = ['condor_status']
        if self.pool:
            query_command.extend(('-pool', str(self.pool)))
        query_command.extend((
            "-startd",
            "-constraint", 'SlotType!="Dynamic"',
            "-af", "TotalSlotCpus", "TotalSlotMemory", "TotalSlotDisk", "Machine"
        ))
        data = {'cpus': 0, 'memory': 0, 'disk': 0, 'machines': 0}
        machines = set()
        for machine_info in subprocess.check_output(query_command, universal_newlines=True).splitlines():
            try:
                cpus, memory, disk, machine = machine_info.split()
            except ValueError:
                continue
            else:
                data['cpus'] += float(cpus)
                data['memory'] += float(memory)
                data['disk'] += float(disk)
                machines.add(machine)
        data['machines'] = len(machines)
        self._valid_date = self.max_age + time.time()
        self._data = data

    def __repr__(self):
        return '%s(pool=%s, max_age=%s)' % (self.__class__.__name__, self.pool, self.max_age)


class PoolResourceView(object):
    def __init__(self, resource, pool_resources):
        self._resource = resource
        self._pool_resources = pool_resources

    def __int__(self):
        return int(self._pool_resources[self._resource])

    def __float__(self):
        return float(self._pool_resources[self._resource])

    def __sub__(self, other):
        return float(self) - other


__all__ = [ConcurrencyConstraintView.__name__, ConcurrencyUsageView.__name__, PoolResourceView.__name__]



__title__ = 'cobald'
__summary__ = 'COBalD - the Opportunistic Balancing Deamon'
__url__ = 'https://github.com/MaineKuehn/cobald'

__version__ = '0.0.1'
__author__ = 'Eileen Kuehn, Max Fischer'
__email__ = 'maxfischer2781@gmail.com'
__copyright__ = '2018 %s' % __author__
__keywords__ = 'opportunistic scheduling scheduler demand feedback loop asyncio'



import argparse
import asyncio
import logging
import platform
import sys

from .pool import ConcurrencyLimit, ConcurrencyAntiLimit

from ...controller.linear import LinearController
from ...proxy.buffer import Buffer
from ...proxy.logger import Logger
from ...proxy.limiter import Limiter
from ...proxy.coarser import Coarser
from ...utility import schedule_pipelines


CLI = argparse.ArgumentParser('HTCondor Concurrency Limit Balancer')
CLI_resource = CLI.add_argument_group('Resource Definition')
CLI_resource.add_argument(
    'resource',
    help='name of the concurrent resource to limit',
    type=str,
)
CLI_resource.add_argument(
    'opponent',
    nargs='?',
    help='name of any concurrent resource that conflicts with `resource`',
)
CLI_resource.add_argument(
    'total',
    nargs='?',
    help='the total underlying resources shared by `resource` and `opponent`. "'
         '"Must be a number or any of "cpus", "memory", "disk" or "machines"',
)
CLI_resource.add_argument(
    '--pool',
    default=None,
    type=str,
    help='name of the condor pool to manage',
)
CLI_controller = CLI.add_argument_group('Resource Control')
CLI_controller.add_argument(
    '--max-rate',
    default=8,
    help='maximum increase or decrease of resources per second',
)
CLI_controller.add_argument(
    '--decrease',
    default=0.9,
    help='utilisation below which resources are decreased',
)
CLI_controller.add_argument(
    '--increase',
    default=0.95,
    help='utilisation above which resources are increased',
)
CLI_controller.add_argument(
    '--limit',
    help='maximum limit for resource',
    type=int,
    default=float('inf')
)
CLI_controller.add_argument(
    '--granularity',
    help='maximum limit for resource',
    default=1,
    type=int,
)


def main():
    event_loop = asyncio.get_event_loop()
    options = CLI.parse_args()
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)-15s (%(process)d) %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    logging.critical('COBalD: condor_limits')
    logging.critical('%s %s (%s)', platform.python_implementation(), platform.python_version(), sys.executable)
    logging.warning('configuring components')
    if options.opponent or options.total:
        if not options.opponent or not options.total:
            raise RuntimeError('both opponent and total must be set together')
        try:
            total = float(options.total)
        except ValueError:
            total = options.total
        pool = ConcurrencyAntiLimit(
            resource=options.resource,
            opponent=options.opponent,
            total=total,
            pool=options.pool,
        )
    else:
        pool = ConcurrencyLimit(resource=options.resource, pool=options.pool)
    pool = Logger(pool, identifier='htcondor.limits')
    pool = Buffer(pool, window=30)
    if options.granularity != 1:
        pool = Coarser(pool, granularity=options.granularity)
    pool = Limiter(pool, minimum=options.granularity, maximum=options.limit)
    pool = LinearController(
        pool,
        low_utilisation=options.decrease, high_consumption=options.increase,
        rate=options.max_rate
    )
    logging.warning('registering components')
    schedule_pipelines(event_loop, pool)
    logging.warning('launching event loop...')
    event_loop.run_forever()


main()



from cobald.interfaces.pool import Pool
from cobald.interfaces.proxy import ProxyPool


class Coarser(ProxyPool):
    """
    Proxy limiting changes in a pool to a certain granularity

    :param target: the pool to which changes are applied
    :param granularity: granularity of changes
    """
    @ProxyPool.demand.setter
    def demand(self, value):
        self.target.demand = value // self.granularity * self.granularity

    def __init__(self, target: Pool, granularity: int = 1):
        super().__init__(target=target)
        self.granularity = granularity



from .pool import Pool
from .controller import Controller


class ProxyPool(Controller, Pool):
    """
    A pool that controls another pool to provide resources
    """
    @property
    def supply(self):
        """The volume of resources that is provided by this site"""
        return self.target.supply

    @property
    def demand(self):
        """The volume of resources to be provided by this site"""
        return self.target.demand

    @demand.setter
    def demand(self, value):
        self.target.demand = value

    @property
    def utilisation(self) -> float:
        """Fraction of the provided resources which is actively used"""
        return self.target.utilisation

    @property
    def consumption(self) -> float:
        """Fraction of the provided resources which is assigned for usage"""
        return self.target.consumption

    def __init__(self, target: Pool):
        super().__init__(target=target)



from typing import Union

from ...interfaces.pool import Pool

from .adapter import ConcurrencyConstraintView, ConcurrencyUsageView, PoolResources


class ConcurrencyLimit(Pool):
    @property
    def supply(self):
        return self._constraints[self.resource]

    @property
    def utilisation(self):
        return self._usage[self.resource] / self._constraints[self.resource]

    consumption = utilisation

    @property
    def demand(self):
        return self._constraints[self.resource]

    @demand.setter
    def demand(self, value):
        self._constraints[self.resource] = value

    def __init__(self, resource: str, pool: str = None):
        self.resource = resource
        self.pool = pool
        self._constraints = ConcurrencyConstraintView(pool=pool)
        self._usage = ConcurrencyUsageView(pool=pool)


class ConcurrencyAntiLimit(Pool):
    @property
    def supply(self):
        return self.total - self._constraints[self.opponent]

    @property
    def utilisation(self):
        return self._usage[self.resource] / self.supply

    consumption = utilisation

    @property
    def demand(self):
        return self.total - self._constraints[self.opponent]

    @demand.setter
    def demand(self, value):
        self._constraints[self.opponent] = self.total - value

    def __init__(self, resource: str, opponent: str, total: Union[str, float], pool: str = None):
        self.resource = resource
        self.opponent = opponent
        self.pool = pool
        if isinstance(total, str):
            pool_resources = PoolResources(pool=pool)
            self.total = pool_resources[total]
        else:
            self.total = total
        self._constraints = ConcurrencyConstraintView(pool=pool)
        self._usage = ConcurrencyUsageView(pool=pool)


__all__ = [ConcurrencyLimit.__name__]



import abc
import asyncio


class Actor(metaclass=abc.ABCMeta):
    """
    An active component that can be run by an event loop
    """
    def mount(self, event_loop: asyncio.AbstractEventLoop):
        """Mount the Actor in ``event_loop`` for execution"""
        event_loop.create_task(self.run())

    @asyncio.coroutine
    def run(self):
        return



import logging

from cobald.interfaces.pool import Pool
from cobald.interfaces.proxy import ProxyPool


class Logger(ProxyPool):
    """
    Proxy which logs all changes to ``target.demand``
    """
    @ProxyPool.demand.setter
    def demand(self, value):
        self._logger.log(
            self.level, 'demand = %s [demand=%s, supply=%s, utilisation=%.2f, consumption=%.2f]',
            value, self.target.demand, self.target.supply, self.target.utilisation, self.target.consumption)
        self.target.demand = value

    @property
    def identifier(self) -> str:
        return self._logger.name

    @identifier.setter
    def identifier(self, value: str):
        if value is None:
            value = self.target.__class__.__qualname__
        self._logger = logging.getLogger(value)

    def __init__(self, target: Pool, identifier: str = None, level: int = logging.INFO):
        super().__init__(target=target)
        self._logger = None  # type: logging.Logger
        self.identifier = identifier
        self.level = level



from cobald.interfaces.pool import Pool
from cobald.interfaces.proxy import ProxyPool


class Limiter(ProxyPool):
    """
    A passive buffer limiting changes to a pool

    :param target: the pool to which changes are applied
    :param minimum: minimum demand to provide
    :param maximum: maximum demand to provide
    """
    @ProxyPool.demand.setter
    def demand(self, value):
        self.target.demand = type(value)(min(self.maximum, max(self.minimum, value)))

    def __init__(self, target: Pool, minimum: float = 1.0, maximum: float = float('inf')):
        super().__init__(target=target)
        self.minimum = minimum
        self.maximum = maximum



import asyncio

from cobald.interfaces.pool import Pool
from cobald.interfaces.controller import Controller
from cobald.interfaces.actor import Actor


class LinearController(Controller, Actor):
    """
    Controller that linearly increases or decreases demand

    :param target: the pool to manage
    :param low_utilisation: pool utilisation below which resources are decreased
    :param high_consumption: pool consumption above which resources are increased
    :param rate: maximum change of demand in resources per second
    """
    @property
    def rate(self):
        return 1 / self._interval

    @rate.setter
    def rate(self, value):
        self._interval = 1 / value

    def __init__(self, target: Pool, low_utilisation=0.5, high_consumption=0.5, rate=1):
        super().__init__(target=target)
        self._interval = None
        self.rate = rate
        self.low_utilisation = low_utilisation
        self.high_consumption = high_consumption

    @asyncio.coroutine
    def run(self):
        while True:
            if self.target.utilisation <= self.low_utilisation:
                self.target.demand -= 1
            elif self.target.consumption >= self.high_consumption:
                self.target.demand += 1
            yield from asyncio.sleep(self._interval)
